@@ -36,7 +36,8 @@ class StreamRangeMerger(RangeMerger2):
 
     # all values for each dimension
     self.all_cont_vals = defaultdict(set) # idx -> values
-    # name -> { val -> [sum, count] }
+
+    # attribute name -> { attr val -> [sum of influence value at c=0.1, count] }
     self.all_disc_vals = defaultdict(lambda: defaultdict(lambda: [0,0])) 
 
     # name -> { val -> # times failed }
@@ -49,7 +50,7 @@ class StreamRangeMerger(RangeMerger2):
     self.adj_graph = None
 
 
-    self.K = 1
+    self.K = 2
     self.nblocks = 50
 
     if len(self.learner.full_table) < 40000:
@@ -58,7 +59,7 @@ class StreamRangeMerger(RangeMerger2):
     if len(self.learner.full_table) < 10000:
       self.nblocks = 100
 
-    self.get_frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks)
+    self.get_frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks, learner=self.learner)
     self.get_frontier.stats = self.stats
 
 
@@ -71,7 +72,7 @@ class StreamRangeMerger(RangeMerger2):
 
   def get_frontier_obj(self, version):
     while version >= len(self.frontiers):
-      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.blocks)
+      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.blocks, learner=self.learner)
       frontier.stats = self.stats
       self.frontiers.append(frontier)
     return self.frontiers[version]
@@ -138,6 +139,7 @@ class StreamRangeMerger(RangeMerger2):
     clusters = set()
     for frontier in self.frontier_iter:
       clusters.update(frontier.seen_clusters)
+    rm_dups(clusters, key=lambda c: str(c.rule))
 
     clusters = sorted(clusters, key=lambda c: c.inf_func(c_val), reverse=True)[:K]
     return clusters
@@ -320,6 +322,8 @@ class StreamRangeMerger(RangeMerger2):
   @instrument
   def update_rejected_directions(self, cluster, dim, direction, val):
     if direction == 'disc':
+      if not hasattr(val, '__iter__'):
+        val = [val]
       for v in list(val):
         self.rejected_disc_vals[dim].append(set([v]))
         self.failed_disc_vals[dim][str(v)] += 1
@@ -332,7 +336,7 @@ class StreamRangeMerger(RangeMerger2):
   def greedy_expansion(self, cluster, seen, version=None, frontier=None):
     _logger.debug("merger\tgreedy_expand\t%s", cluster.rule.simplify())
     if frontier is None:
-      frontier = CheapFrontier(self.c_range, K=1, nblocks=15)
+      frontier = CheapFrontier(self.c_range, K=1, nblocks=15, learner=self.learner)
       frontier.stats = self.stats
       frontier.update([cluster])
 
@@ -379,6 +383,13 @@ class StreamRangeMerger(RangeMerger2):
 
 
 class PartitionedStreamRangeMerger(StreamRangeMerger):
+  """
+  Partitions the merger based on user defined labels so that frontier curves from 
+  one partition do not suppress curves in another partition
+
+  MR labels based on dimensionality
+  BDT labels as leaf/non-leaf
+  """
 
   def __init__(self, *args, **kwargs):
     super(PartitionedStreamRangeMerger, self).__init__(*args, **kwargs)
@@ -388,7 +399,7 @@ class PartitionedStreamRangeMerger(StreamRangeMerger):
   def get_frontier_obj(self, version, partitionkey):
     frontiers = self.frontiers[partitionkey]
     while version >= len(frontiers):
-      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks)
+      frontier = CheapFrontier(self.c_range, K=self.K, nblocks=self.nblocks, learner=self.learner)
       frontier.stats = self.stats
       frontiers.append(frontier)
     return frontiers[version]
@@ -422,6 +433,13 @@ class PartitionedStreamRangeMerger(StreamRangeMerger):
     if not skip_frontier:
       clusters, _ = frontier.update(clusters)
 
+    # XXX: new cluster should be better than _all_ frontiers
+    #for f in self.frontier_iter:
+    #clusters, _ = f.update(clusters)
+
+    if not clusters: 
+      return clusters
+
     if self.DEBUG and not skip_frontier:
       print "base_frontier"
       self.print_clusters(clusters)
